"""Data Preparation class to handle data processing for machine learning."""

__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2019-10-16'

import datetime as dt
import inspect
import logging
from typing import Union, List

import pandas as pd
import xarray as xr

from mlair import helpers
from mlair.helpers import join
from mlair.data_handling.data_preparation import AbstractDataPrep

# define a more general date type for type hinting
date = Union[dt.date, dt.datetime]
str_or_list = Union[str, List[str]]
number = Union[float, int]
num_or_list = Union[number, List[number]]
data_or_none = Union[xr.DataArray, None]


class DataPrepJoin(AbstractDataPrep):
    """
    This class prepares data to be used in neural networks.

    The instance searches for local stored data, that meet the given demands. If no local data is found, the DataPrep
    instance will load data from TOAR database and store this data locally to use the next time. For the moment, there
    is only support for daily aggregated time series. The aggregation can be set manually and differ for each variable.

    After data loading, different data pre-processing steps can be executed to prepare the data for further
    applications. Especially the following methods can be used for the pre-processing step:

    - interpolate: interpolate between data points by using xarray's interpolation method
    - standardise: standardise data to mean=1 and std=1, centralise to mean=0, additional methods like normalise on \
        interval [0, 1] are not implemented yet.
    - make window history: represent the history (time steps before) for training/ testing; X
    - make labels: create target vector with given leading time steps for training/ testing; y
    - remove Nans jointly from desired input and output, only keeps time steps where no NaNs are present in X AND y. \
        Use this method after the creation of the window history and labels to clean up the data cube.

    To create a DataPrep instance, it is needed to specify the stations by id (e.g. "DEBW107"), its network (e.g. UBA,
    "Umweltbundesamt") and the variables to use. Further options can be set in the instance.

    * `statistics_per_var`: define a specific statistic to extract from the TOAR database for each variable.
    * `start`: define a start date for the data cube creation. Default: Use the first entry in time series
    * `end`: set the end date for the data cube. Default: Use last date in time series.
    * `store_data_locally`: store recently downloaded data on local disk. Default: True
    * set further parameters for xarray's interpolation methods to modify the interpolation scheme

    """

    def __init__(self, path: str, station: Union[str, List[str]], variables: List[str], network: str = None,
                 station_type: str = None, **kwargs):
        self.network = network
        self.station_type = station_type
        params = helpers.remove_items(inspect.getfullargspec(AbstractDataPrep.__init__).args, "self")
        kwargs = {**{k: v for k, v in locals().items() if k in params and v is not None}, **kwargs}
        super().__init__(**kwargs)

    def download_data(self, file_name, meta_file):
        """
        Download data and meta from join.

        :param file_name: name of file to save data to (containing full path)
        :param meta_file: name of the meta data file (also containing full path)
        """
        data, meta = self.download_data_from_join(file_name, meta_file)
        return data, meta

    def check_station_meta(self):
        """
        Search for the entries in meta data and compare the value with the requested values.

        Will raise a FileNotFoundError if the values mismatch.
        """
        if self.station_type is not None:
            check_dict = {"station_type": self.station_type, "network_name": self.network}
            for (k, v) in check_dict.items():
                if v is None:
                    continue
                if self.meta.at[k, self.station[0]] != v:
                    logging.debug(f"meta data does not agree with given request for {k}: {v} (requested) != "
                                  f"{self.meta.at[k, self.station[0]]} (local). Raise FileNotFoundError to trigger new "
                                  f"grapping from web.")
                    raise FileNotFoundError

    def download_data_from_join(self, file_name: str, meta_file: str) -> [xr.DataArray, pd.DataFrame]:
        """
        Download data from TOAR database using the JOIN interface.

        Data is transformed to a xarray dataset. If class attribute store_data_locally is true, data is additionally
        stored locally using given names for file and meta file.

        :param file_name: name of file to save data to (containing full path)
        :param meta_file: name of the meta data file (also containing full path)

        :return: downloaded data and its meta data
        """
        df_all = {}
        df, meta = join.download_join(station_name=self.station, stat_var=self.statistics_per_var,
                                      station_type=self.station_type, network_name=self.network, sampling=self.sampling)
        df_all[self.station[0]] = df
        # convert df_all to xarray
        xarr = {k: xr.DataArray(v, dims=['datetime', 'variables']) for k, v in df_all.items()}
        xarr = xr.Dataset(xarr).to_array(dim='Stations')
        if self.kwargs.get('store_data_locally', True):
            # save locally as nc/csv file
            xarr.to_netcdf(path=file_name)
            meta.to_csv(meta_file)
        return xarr, meta

    def __repr__(self):
        """Represent class attributes."""
        return f"Dataprep(path='{self.path}', network='{self.network}', station={self.station}, " \
               f"variables={self.variables}, station_type={self.station_type}, **{self.kwargs})"


if __name__ == "__main__":
    dp = DataPrepJoin('data/', 'dummy', 'DEBW107', ['o3', 'temp'], statistics_per_var={'o3': 'dma8eu', 'temp': 'maximum'})
    print(dp)
